%%%============================================================================
%%% Time-stamp: <2007-09-05 20:07:09 j>
%%%
%%% File    : burm.hrl
%%% Purpose : BURM support procedures.
%%% Author  : j@crunchy.svirfneblin.org
%%% License : BSD
%%% Created : 2007-08-30 09:58:20 j
%%%
%%%============================================================================

-define(DEBUG, true).

%% For conditional tracing
-ifdef(DEBUG).
-define(dbg(Fmt, Data), io:format(Fmt, Data)).
-else.
-define(dbg(Fmt, Data), true).
-endif.

%%%_. To-do list

%% TODO: describe interface bw spec and burm
%%         o BURM exports: rewrite/1, label/1, init/0, reduce/2, etc,
%%                         INN, IRN macros
%%         o BURM requires: op_label/1, kids/1, kids_set/2, state_label/1
%%                          state_label_set/2
%% TODO: describe types (node, cost, rule, result)
%% TODO: check names in funs which include user code (state, closure, action)
%%         o BurmFoo, burm_, etc
%% TODO: exports

%%%_. State labels

-include("eburg_shared.hrl").

%% ?burm_cost(Nt: int, Node: node) -> int
%% returns the cost of deriving `Nt' at `Node'.
-define(burm_cost(Nt, Node),
        element(Nt, (state_label(Node))#burm_state.cost)).

%% ?burm_rule(Nt: int, Node: node) -> int
%% returns the optimal rule for deriving `Nt' at `Node'.
-define(burm_rule(Nt, Node),
        element(Nt, (state_label(Node))#burm_state.rule)).

%%%_. Labeler & reducer

%% burm_maybe_init() -> ok
%% calls `init' if it has not yet been called.
%% FIXME
burm_maybe_init() ->
    case get('burm_init_done') of
        undefined ->
            put('burm_init_done', true),
            init();
        _ ->
            ok
    end.

%% label(Tree: node) -> node
%% returns `Tree' with the `state_label' field set.
label(Tree0) ->
    Op = op_label(Tree0),
    ?dbg("~p: entering label/1~n", [Op]),

    ?dbg("~p: labeling kids~n", [Op]),
    Kids0  = kids(Tree0),
    Kids1  = lists:map(fun label/1, Kids0), % XXX: pmap?
    Tree1  = kids_set(Tree0, Kids1),
    Labels = lists:map(fun state_label/1, Kids1),
    ?dbg("~p: kids labeled~n", [Op]),
    
    ?dbg("~p: checking state table~n", [Op]),
    case ets:lookup(burm_labels, {Op, Labels}) of
        [] ->
            ?dbg("~p: no label found, calculating~n", [Op]),
            Label = burm_state(Tree1),
            ets:insert(burm_labels, {{Op, Labels}, Label}),
            ?dbg("~p: label calculated and saved to state table -- ~p~n~n",
                 [Op, Label]);
        [{_, Label}] ->
            ?dbg("~p: label found -- ~p~n~n", [Op, Label])
    end,
    state_label_set(Tree1, Label).

%% reduce(Tree: node, Nt: int) -> result
%% returns the result of executing the user-supplied semantic actions
%% associated with `Tree'.
reduce(Tree, Nt) ->
    Op    = op_label(Tree),
    Rule  = ?burm_rule(Nt, Tree),
    Kids0 = kids(Tree),
    Nts   = nts(Rule),
    ?dbg("~p: reducing to ~s~n", [Op, burm_string({nt, Nt})]),
    ?dbg("~p: rule is ~s~n", [Op, burm_string({rule, Rule})]),
    case Rule of
        x ->
            %% XXX will throw nomatch here once normalization is implemented
            reduce(lists:nth(1, Kids0), Nt);
        _ -> case {length(Kids0), length(Nts)} of
                 {0, 1} -> 
                     ?dbg("~p: chainrule, recursing~n", [Op]),
                     %% FIXME is_chainrule, ugh!
                     %% when is_chainrule, call again with new nt on same tree
                     KX = reduce(Tree, hd(Nts)),
                     ?dbg("~p: firing action~n~n", [Op]),
                     burm_action(Tree, [KX], Rule);
                 {N, N} ->
                     ?dbg("~p: regular rule, reducing kids~n", [Op]),
                     Kids1 = [reduce(T, N) ||
                                 {T, N} <- lists:zip(Kids0, nts(Rule))],
                     ?dbg("~p: firing action~n~n", [Op]),
                     burm_action(Tree, Kids1, Rule)
             end
    end.

%%%_. Error handling

%% burm_assert(Expr: bool) -> ok
%% returns `ok' iff its argument is true.
burm_assert(true)  -> ok;
burm_assert(false) -> burm_panic(assert).

%% burm_panic(Reason: error_descriptor)
%% throws a `burm_exn' exception.
burm_panic(Reason) -> throw({burm_exn, Reason}).

%% format_error(Reason: error_descriptor) -> string
%% returns a descriptive error message for `Reason'.
format_error(assert) ->
    "assertion failed";
format_error(badopts) ->
    "`rewrite' called with bad options";
format_error(nomatch) ->
    "subject tree cannot be matched".

%%%_. Debugging aids

%% dumpcover(Tree: node, Nt: int, Lvl: int) -> ok
%% returns `ok' after printing the `Tree's rule cover.
dumpcover(Tree, Nt, Lvl) ->
    Rule = ?burm_rule(Nt, Tree),
    Nts  = nts(Rule),
    Kids = kids(Tree),
    if Rule == x -> dumpcover(lists:nth(1, Kids), Nt, Lvl); % XXX
       true ->
            case is_chainrule(Rule) of
                true ->
                    burm_print("~s ~n", [burm_string({rule, Rule})], Lvl),
                    dumpcover(Tree, hd(Nts), Lvl+1);
                false ->
                    burm_print("~s ~n", [burm_string({rule, Rule})], Lvl),
                    [dumpcover(T, N, Lvl+1) || {T, N} <- lists:zip(Kids, Nts)]
            end
    end,
    ok.

%% dumplabels(Tree: node) -> ok
%% returns `ok' after printing `Tree's state labels.
dumplabels(Tree) ->
    io:put_chars("Operator: Nonterminal using Rule for Cost\n"),
    io:put_chars("=========================================\n"),
    dumplabels(Tree, 0).

dumplabels(Tree, Lvl) ->
    #burm_state{cost=C, rule=R} = state_label(Tree),
    burm_print("~p:~n", [op_label(Tree)], Lvl),
    [burm_print("~s using ``~s'' for ~p~n",
                [burm_string({nt,N}), burm_string({rule,element(N,R)}),element(N,C)],
                Lvl+1)
     || N <- lists:seq(1, length(tuple_to_list(C))),
        element(N, R) /= x],
    [dumplabels(Tree, Lvl+2) || Tree <- kids(Tree)],
    ok.

%% burm_print(Fmt: string, Data: [term], N: int) -> ok
%% returns `ok' after `format'ting `Data' and indenting it `N' levels.
burm_print(Fmt, Data, Lvl) ->
    io:format(lists:duplicate(Lvl*2, $\s)++Fmt, Data).

%%%_. Emacs

%%% Local Variables:
%%% mode: erlang
%%% allout-layout: t
%%% End:
