%%%============================================================================
%%% Time-stamp: <2007-09-06 10:09:05 j>
%%%
%%% File    : eburg.erl
%%% Purpose : A code generator generator for Erlang.
%%% Author  : j@crunchy.svirfneblin.org
%%% License : BSD
%%% Created : 2007-07-22 15:49:04 j
%%%
%%%============================================================================

%% @author Jakob Sievers <j@crunchy.svirfneblin.org>
%% @copyright 2007 Jakob Sievers
%% @version 0.1

%% @doc
%% An iburg-style code generator generator for Erlang.
%% ...
%% @end

%% @see //yecc and //leex

%% == Related Software ==
%% iburg is available from [http://www.cs.princeton.edu/software/iburg/],
%% ML-Burg is part of the Standard ML of New Jersey distribution
%% ([http://www.smlnj.org/]), and OcamlBurg is hosted on Google Code at
%% [http://code.google.com/p/ocaml-burg/].

%% == Papers ==
%% @reference Christopher W. Fraser, David R. Hanson, and Todd A. Proebsting.
%% <em>Engineering a simple, efficient code generator generator</em>.
%% AMC Letters on Programming Languages and Systems, 1992.

%% @reference Christopher Fraser, David Hanson.
%% <em>A Retargetable C Compiler: Design and Implementation</em>.
%% Addison-Wesley, 1995.

%% @reference M. Anton Ertl, Kevin Casey, David Gregg.
%% <em>Fast and Flexible Instruction Selection with On-Demand Tree-Parsing
%% Automata</em>.
%% ACM Conference on Programming Language Design and Implementation, 2006.

%%%_. To-do list

%% TODO: Write user manual
%% TODO: normalization & state equivalence
%% TODO: check dyncosts, improve dyncost handling throughout, test dyncosts
%% TODO: write test spec for special cases
%% TODO: better error messages (line numbers &c.)
%% TODO: collect errors and warnings (dont abort on first error)
%% TODO: more support for user written reducers?
%% TODO: option to dump burm_labels ets table
%% TODO: parallelize labeler?
%% TODO: better error handling in BURM...
%% TODO: large specs & subject trees -> lcc x86 grammar, gforth grammar?
%% TODO: get lcc to dump ir trees?
%% TODO: use eunit for test suite
%% TODO: check for loops in rules
%% TODO: rewrite nonterms, arities, etc using patmap

%%%_. Declarations

-module(eburg).
-author('j@crunchy.svirfneblin.org').

-export([main/1, file/1, make/0]).

-import(lists, [reverse/1, delete/2, member/2, append/1, duplicate/2]).
-import(lists, [seq/2, zip/2, last/1, flatten/1, usort/1]).
-import(lists, [map/2, filter/2, foreach/2, flatmap/2, all/2, foldl/3]).

-compile([export_all]).
-include("eburg.hrl").
-include("eburg_shared.hrl").

%%%_. Interface

%% @spec main() -> ok | error
%% @doc For command-line use.
%% @equiv file(File)
main([File]) ->
    file(File);
main(_) ->
    usage().

%% @spec file(File::name()) -> ok | error
%% where
%%       name() = file:name()
%% @doc returns `error' and prints one or more error messages when BURM couldn't
%% be generated from the specification in `File', otherwise, it returns `ok'
%% after writing the generated program to ```filename:basename(File)++".erl"'''.
file(File) ->
    try
        emit(check(read(File)))
    catch
        throw:
          {eburg_error, File, Line, Reason} ->
            err(File, Line, format_error(Reason)),
            error;
          {eburg_error, File, Reason} ->
            err(File, format_error(Reason)),
            error
    end.

%% usage()
%% prints an error message and exits.
usage() ->
    io:format("usage: eburg file~n"),
    halt(1).

%% @spec make() -> ok
%% @doc returns `ok' after generating and compiling the lexer and parser for
%% eburg specification files.
make() ->
    leex:file(eburg_lexer),
    c:c(eburg_lexer),
    yecc:file(eburg_parser),
    c:c(eburg_parser).

%%%_. Code

%%%_ , Read specification file

%%%_  . read/1

%%-----------------------------------------------------------------------------
%% `read' first reads the input file into memory and uses a leex/yecc generated
%% lexer/parser to construct an #eburg record. It then adds some information
%% derived from the input spec, notably:
%%   o nts, the known nonterminals
%%   o crules, chainrules indexed by RHS nonterminal
%%   o oprules, regular rules indexed by RHS operator
%%   o nt_nums, mapping of nonterminals to Internal Nonterminal Numbers (INNs)
%%   o arities, mapping of operators to their arities
%%   o Internal Rule Numbers (IRNs)
%% INNs and IRNs number nonterminals/rules _in the order in which they appear
%% in the spec file_.
%%-----------------------------------------------------------------------------

%% read(File: name) -> #eburg
%% returns the initial state record.
read(File) ->
    Dir     = filename:dirname(File) ++ "/",
    Module  = filename:basename(File, ".brl"),
    InFile  = Dir ++ Module ++ ".brl",
    OutFile = Dir ++ Module ++ ".erl",
    S0      = parse(InFile),
    S1      = reparse(S0),
    S1#eburg{in_file=InFile, out_file=OutFile, module=Module}.

%%%_  . Character I/O and parsing

%% parse(File: string) -> #eburg
%% returns the internal representation of the specification in `File'.
%% throws `file' if `File' couldn't be opened.
%% throws `syntax' if the specification contains a syntax error.
parse(File) ->
    case file:open(File, read) of
        {error, Reason} ->
            die(File, {file, Reason});
        {ok, Stream} ->
            try
                {Spec, Code} = get_file(Stream),
                Reader       = mk_reader(fun eburg_lexer:lex/1,
                                         fun eburg_parser:parse/1),
                case Reader(Spec) of
                    {error, Reason, Line} -> die(File, Line, {syntax, Reason});
                    S             -> S#eburg{code=Code}
                end
             after
                 file:close(Stream)
             end
    end.

%% get_file(File: io_device) -> {string, string}
%% returns the contents of `File', split into declarations & rules
%% and user code.
get_file(File) ->
    get_file(File, []).

get_file(File, Lines) ->
    case io:get_line(File, '') of
        eof ->
            {Lines, ""};
        "Erlang code.\n" ->
            {Rest, ""} = get_file(File), % assuming one ``Erlang code.\n''
            {Lines, Rest};
        Line ->
            get_file(File, Lines ++ Line)
    end.

%%%_  . Add convenience data structures

%% reparse(S: #eburg) -> #eburg
%% returns `S' with blanks filled in.
reparse(S) ->
    Rs  = [R#rule{number=N} || {R, N} <- enumerate(S#eburg.rules)],
    Nts = uniq([R#rule.lhs || R <- Rs]),
    S#eburg{nts      = Nts,
            arities  = usort(arities([R#rule.rhs || R <- Rs])), % checked later
            nt_nums  = enumerate(Nts),
            rules    = Rs,
            crules   = filter_and_group_by(Nts, Rs),
            oprules  = filter_and_group_by(S#eburg.ops, Rs)}.

%% enumerate(Lst: [term]) -> [{term, int}]
%% returns `Lst' with sequence numbers attached to each of its elements.
enumerate(Lst) -> zip(Lst, seq(1, length(Lst))).

%% uniq(Lst: [term]) -> [term]
%% returns `Lst' with duplicate elements removed (first occurence kept).
uniq(Lst) -> uniq(Lst, []).

uniq([], Ys) ->
    Ys;
uniq([X|Xs], Ys) ->
    case member(X, Ys) of
        true  -> uniq(Xs, Ys);
        false -> uniq(Xs, Ys ++ [X])
    end.

%% arities(Ps: [#pattern]) -> [{atom, int}]
%% returns the operators which occur in `Ps' paired with their arities.
arities([])                            -> [];
arities([P|Ps])                        -> arities(P) ++ arities(Ps);
arities(#pattern{op=O,arity=N,kids=K}) -> [{O, N}] ++ arities(K).

%% filter_and_group_by(Keys: [atom], Rs: [#rule]) -> [{atom, [#rule]}]
%% returns those `Rs' with a root operator in `Keys', grouped by root operator.
filter_and_group_by(Keys, Rs) ->
    [{Key, [R || R <- Rs, (R#rule.rhs)#pattern.op == Key]} || Key <- Keys].

%%%_ , Basic input validation

%%%_  . check/1

%%-----------------------------------------------------------------------------
%% `check' tests for the following:
%%   o syntactic correctness of semantic actions and user code (using stdlib
%%     facilities)
%%   o sanity of declarations
%%   o validity of tree patterns
%%   o all operators used, all nonterminals reachable
%%-----------------------------------------------------------------------------

%% check(S: #eburg) -> #eburg
%% returns its input unchanged if it is found to be valid.
check(S) ->
%% XXX
%%      check_actions(S),
%%      check_code(S),

    check_uniq(S),
    check_diff(S),
    check_start(S),

    check_rhs(S),
    check_arities(S),

    case catch check_used(S) of
        {eburg_warning, Reason0} ->
            err(S#eburg.in_file, "Warning: " ++ format_error(Reason0));
        _ -> ok
    end,
    case catch check_reachability(S) of
        {eburg_warning, Reason1} ->
            err(S#eburg.in_file, "Warning: " ++ format_error(Reason1));
        _ -> ok
    end,

    S.

%%%_  . Erlang code

%% check_actions(S: #eburg) -> ok
%% returns `ok' if all the semantic actions can be parsed.
%% throws `action' if they can't.
check_actions(S) ->
    Reader = mk_reader(fun erl_scan:string/1, fun erl_parse:parse_exprs/1),
    foreach(fun ({R, {error, _Reason, _Line}}) ->
                    die(S#eburg.in_file, R#rule.line, action);
                ({_, _}) -> ok
            end,
            map(fun (R) -> {R, Reader(R#rule.action)} end,
                S#eburg.rules)).

%% check_code(S: #eburg) -> ok
%% returns `ok' if the user code is syntactically valid.
%% throws `code' if it's not.
%% FIXME: code offset in lines?
check_code(#eburg{code=""}) ->
    ok;
check_code(S) ->
    Reader = mk_reader(fun erl_scan:string/1, fun erl_parse:parse_exprs/1),
    case Reader(S#eburg.code) of
        {error, _Reason, Line} ->
            die(S#eburg.in_file, Line, code);
        _ -> ok
    end.

%%%_  . Declarations

%% check_uniq(S: #eburg) -> true
%% returns `true' iff no operator has been declared twice.
%% throws `uniq' if there is at least one which has.
check_uniq(S) ->
    (length(S#eburg.ops) == length(usort(S#eburg.ops)))
        orelse die(S#eburg.in_file, uniq).

%% check(S: #eburg) -> true
%% returns `true' iff the operator and nonterminal names are disjoint.
%% throws `diff' otherwise.
check_diff(S) ->
    all(fun (Op) -> not member(Op, S#eburg.nts) end, S#eburg.ops)
        orelse die(S#eburg.in_file, diff).

%% check_start(S: #eburg) -> ok
%% returns `ok' iff S#eburg.start is a known nonterminal.
%% throws `start' otherwise.
check_start(S) ->
    case member(S#eburg.start, S#eburg.nts) of
        true  -> ok;
        false -> die(S#eburg.in_file, start)
    end.

%%%_  . Tree patterns

%% check_rhs(S: #eburg) -> ok
%% returns `ok' if all right-hand sides are valid, i.e.
%%   (a) #pattern.op fields hold only declared operators or nonterminals
%%   (b) only declared operators have kids.
%% throws `undeclared' otherwise.
check_rhs(S) ->
    foreach(fun (ok) -> ok;
                ({error, Line, Op}) ->
                    die(S#eburg.in_file, Line, {undeclared, Op})
            end,
            map(fun (R) -> catch check_pats(R, S) end, S#eburg.rules)).

%% check_pats(R: #rule, S: #eburg) -> ok
%% returns `ok' if `R's RHS is valid.
%% throws {error, Line, Op} if `Op' is an undeclared operator which occurs on
%% the RHS of `R'. `Line' is `R's line number.
check_pats(R, S) ->
    check_pats([R#rule.rhs], R, S).

check_pats([], _R, _S) ->
    ok;
check_pats([#pattern{op=Op, kids=[]}|Ps], R, S) -> % (a)
    case member(Op, S#eburg.ops) orelse member(Op, S#eburg.nts) of
        true  -> check_pats(Ps, R, S);
        false -> throw({error, R#rule.line, Op})
    end;
check_pats([#pattern{op=Op, kids=Kids}|Ps], R, S) -> % (b)
    case member(Op, S#eburg.ops) of
        true  ->
            check_pats(Kids, R, S),
            check_pats(Ps, R, S);
        false ->
            throw({error, R#rule.line, Op})
    end.

%% check_arities(S: #eburg) -> true
%% returns `true' if no operator is used in multiple places with different
%% arities
%% throws `arity' otherwise.
check_arities(S) ->
    all(fun (1) -> true; (_) -> false end,
        [length(proplists:get_all_values(Op, S#eburg.arities)) ||
            Op <- S#eburg.ops]) orelse die(S#eburg.in_file, arity).

%%%_  . Consistency

%% check_used(S: #eburg) -> true
%% returns `true' if all operators appear on the RHS of some rule.
check_used(S) ->
    explore([R#rule.rhs || R <- S#eburg.rules],
            S#eburg.ops,
            fun (P) -> P#pattern.kids end,
            unused).

%% check_reachability(S: #eburg) -> true
%% returns `true' if all nonterminals are reachable from the start symbol.
%% Inspired by ML-Burg.
check_reachability(S) ->
    Rs = S#eburg.rules,
    explore([R#rule.rhs || R <- Rs, R#rule.lhs == S#eburg.start],
            S#eburg.nts,
            fun (P) -> [R#rule.rhs || R <- Rs, R#rule.lhs == P#pattern.op] end,
            unreachable).

%% explore(Init: [#pattern], Ops: [atom], Step: fun, Tag: atom) -> true
%% returns `true' if all `Ops' are `reach'able from `Init' via `Step'.
%% throws `Tag' otherwise.
explore(Init, Ops0, Step, Tag) ->
    Ops1   = sets:from_list(Ops0),
    Seen   = reach(Init, Ops1, Step),
    Unseen = sets:subtract(Ops1, Seen),
    sets:size(Unseen) == 0 orelse warning('FIXME', {Tag, sets:to_list(Unseen)}).

%% reach(Pats: [#pattern], Ops: set, Step: fun) -> set
%% returns the elements of `Ops' which occur on the RHS of some element of
%% `Pats'. `Step' applied to a single pattern should return a list of patterns.
reach(Pats, Ops, Step) -> reach(Pats, Ops, Step, sets:new()).

reach([], _, _, Acc) ->
    Acc;
reach([P = #pattern{op=Op}|Ps], Ops, Step, Acc0) ->
    case sets:is_element(Op, Ops) of
        true ->
            %% we've now seen Op
            Acc1 = sets:add_element(Op, Acc0),
            %% expand the search (include patterns one Step from Op)
            Acc2 = sets:union(Acc1, reach(Step(P), Ops, Step)),
            %% process the remaining patterns from the original invocation
            reach(Ps, Ops, Step, Acc2);
        false ->
            %% we're not interested, just continue
            reach(Ps, Ops, Step, Acc0)
    end.

%%%_ , Emit BURM

%%%_  . emit/1

%%-----------------------------------------------------------------------------
%% Overview
%% ========
%% Most of BURM is printed directly to the output file using `io:format'
%% (which makes this process look somewhat quasiquotation-like).
%%
%% The convention is that `emit_*' routines print to ```get('$out_stream')'''
%% (via `say') whereas `gen_*' routines return code (strings) for inclusion in
%% some `emit_*' routine.
%%
%% The following optimizations from [FRASER] are currently implemented:
%%   o inline record
%%   o closure routines
%%   o precompute leaf states (by initializing the state lookup table)
%%
%% There are two main parts which vary from BURM to BURM:
%%   o The labeler (`init', `burm_state' and `burm_closure_*')
%%   o The reducer (`nts', `burm_action')
%% Additionally, there is a preamble which depends slightly on the input
%% specification.
%% Everything else is provided by ``burm.hrl''.
%%
%% A note on INNs and IRNs
%% =======================
%% In addition to being stored in the state record (see `reparse'), the
%% generated source files includes a series of macro definitions mapping
%% human readable names to internal numbers.
%%
%% The numbers are of significance when interpreting a node's state label's
%% rule and cost vectors:
%% element(?NONTERMINAL, (state_label(Node))#burm_state.rule) returns
%% the IRN of the optimal rule for rewriting Node to nonterminal.
%% This IRN is then used to obtain the INNs of the nonterminals occuring on the
%% RHS of the rule by passing it to `nts'.
%%-----------------------------------------------------------------------------

%% emit(S: #eburg) -> ok
%% returns `ok'. Opens the output stream used by `say' and calls `emit_burm'.
%% throws `file' if the output file couldn't be opened.
emit(S) ->
    case file:open(S#eburg.out_file, write) of
        {error, Reason} ->
            die(S#eburg.out_file, {file, Reason});
        {ok, Stream}    ->
            try
                put('$out_stream', Stream),
                emit_burm(S),
                pp(S#eburg.out_file)
            after
                file:close(Stream)
            end
    end.

%% emit_burm(S: #eburg) -> ok
%% returns `ok' after printing BURM.
emit_burm(S) ->
    say("%%% Generated by eburg, do not modify."),
    emit_preamble(S),
    emit_labeler(S),
    emit_reducers(S),
    say("%%% eof").

%%%_  . Preamble

%%-----------------------------------------------------------------------------
%% The preamble consists of compiler declarations, a few macros, user code, and
%% the interface to the generated code generator.
%%-----------------------------------------------------------------------------

%% emit_preamble(S: #eburg) -> ok
%% returns `ok' after printing the BURM preamble.
emit_preamble(S) ->
    %% compiler stuff
    say("-module(~s).", [S#eburg.module]),
    say("-export([init/0, label/1, nts/1, rewrite/1, dumpcover/1])."), % XXX
    %% INNs
    foreach(fun ({Nt, N}) -> say("-define(~s, ~p).", [Nt, N]) end,
            [{to_string({macdef, Nt}), N} || {Nt,N} <- enumerate(S#eburg.nts)]),
    %% IRNs
    foreach(fun({R, N}) -> say("-define(~s, ~p).", [R, N]) end,
            [{to_string({macdef, R}), R#rule.number} || R <- S#eburg.rules]),

    %% user code
    emit_ruler(),
    io:put_chars(get('$out_stream'), (S#eburg.code)),
    emit_ruler(),
    say(""),

    %% exports
    say("-include(\"burm.hrl\")."),
    say("%% rewrite(Tree: tree) -> result"),
    say("%% returns the result of labeling and reducing tree."),
    say("rewrite(Tree) ->"),
    say("    %burm_maybe_init(),"),
    say("    reduce(label(Tree), ~s).", [to_string({macref, S#eburg.start})]),
    say("%% dumpcover(Tree: node) -> ok"),
    say("%% returns `ok' after printing the rule cover for `Tree'."),
    say("dumpcover(Tree) ->"),
    say("    dumpcover(label(Tree), ~s, 0).", [to_string({macref, S#eburg.start})]).

%% emit_ruler() -> ok
%% returns `ok' after printing a dashed comment line.
emit_ruler() ->
    say("%%~s", [duplicate(77, $-)]).

%%%_  . Labeler

%%%_   , Skeleton

%%-----------------------------------------------------------------------------
%% The labeler consists of two parts: `init', which initializes the state
%% lookup table at compile-compile time, and `label' which uses
%% `burm_state' and `burm_closure_*' to compute state labels at compile time.
%% `label' is hardcoded whereas `init', `burm_state', and the
%% `burm_closure_*' routines have to be generated.
%%-----------------------------------------------------------------------------

%% emit_labeler(S: #eburg) -> ok
%% returns `ok' after printing the functions needed for labeling a subject
%% tree.
emit_labeler(S) ->
    emit_init(S),
    emit_state(S),
    emit_closures(S).

%%%_   , init/0

%%-----------------------------------------------------------------------------
%% `emit_init' simulates the two-stage (recording non-chainrules and computing
%% the chainrule closure) labelling process for leaf nodes.
%%-----------------------------------------------------------------------------

%% emit_init(S: #eburg) -> ok
%% returns `ok' after printing code for initializing the state table.
%% XXX annotations!
emit_init(S) ->
    say("%% init() -> ok"),
    say("%% returns `ok' after initializing the state table."),
    say("init() ->"),
    say("ets:new(burm_labels, [named_table]),"),
    sayforeach(fun ({Leaf, Label}) ->
                       say("ets:insert(burm_labels, {{~s, []}, ~p})",
                           [Leaf, Label])
               end,
               ",",
               sim_leaves(S)),
    say(".").

%% sim_leaves(S: #eburg) -> [{atom, #burm_state}]
%% returns the leaf operators paired with their state labels.
sim_leaves(S) ->
    [sim_leaf({L, Rs}, S) ||
        {L, Rs} <- [{Leaf, [R || R <- S#eburg.rules,
                                 (R#rule.rhs)#pattern.op == Leaf,
                                 %% it's a leaf
                                 (R#rule.rhs)#pattern.arity == 0,
                                 %% static cost
                                 is_integer(R#rule.cost)]} ||
                       Leaf <- S#eburg.ops],
        Rs /= []].

%% sim_leaf({Leaf, Rules}: {atom, [#rule]}, S: #eburg) -> #burm_state
%% returns the state label for `Leaf'.
sim_leaf({Leaf, Rules}, S) ->
    Init = list_to_tuple(duplicate(length(S#eburg.nts), x)),
    {Leaf,
     foldl(fun (Rule, State) -> sim_rule(Rule, State, S) end,
           #burm_state{cost=Init, rule=Init},
           Rules)}.

%% sim_rule(Rule: #rule, State: #burm_state, S: #eburg) -> #burm_state
%% returns the result of recording `Rule' in `State'.
sim_rule(Rule, State, S) ->
    sim_record(Rule, State, 0, S).

%% sim_closure(Nt: atom, St: #burm_state. B0: int, S: #eburg) -> #burm_state
%% returns the closure of `St' under `Nt'.
%% XXX Closure when chain rule has dyn cost!
sim_closure(Nt, St, B0, S) ->
    foldl(fun (Rule, State) -> sim_record(Rule, State, B0, S) end,
          St,
          assoc(Nt, S#eburg.crules)).

%% sim_record(R: #rule, State: #burm_state, B0: int, S: #eburg) -> #burm_state
%% returns the result of recording `R' in `State'.
sim_record(#rule{number=N, lhs=Nt, cost=Cost}, State, B0, S) ->
    C0 = State#burm_state.cost,
    R0 = State#burm_state.rule,
    B1 = B0 + Cost, % known to be integer
    M  = assoc(Nt, S#eburg.nt_nums),
    case B1 < element(M, C0) of
        true ->
            C1 = setelement(M, C0, B1),
            R1 = setelement(M, R0, N),
            sim_closure(Nt, #burm_state{cost=C1, rule=R1}, B1, S);
        false ->
            State
    end.

%%%_   , state, burm_closure_*

%%%_    . label updates

%%-----------------------------------------------------------------------------
%% The `burm_state' and `burm_closure_*' routines interact with subject trees
%% in two ways, namely:
%%  (a) By accessing the current node's children.
%%  (b) By updating the current node's state label.
%%
%% For the routines which generate code to do (a), see section
%% ``tree access''.
%% (b) is slightly trickier because intermediate versions of a node's state
%% label have to be named by different temporary variables and the names of
%% those variables have to be communicated to subsequent invocations of the
%% generation routines.
%%
%% The current convention is to store the cost part of a state label in a
%% variable named CostN (where N is some integer), the rule part in a variable
%% named RuleN, and the base cost (sum of child costs inside `burm_state',
%% pre-closure cost inside `burm_closure_*') in a variable named BaseN.
%%
%% FIXME describe temp var name handling
%%-----------------------------------------------------------------------------

%% emit_state(S: #eburg) -> ok
%% returns `ok' after printing BURM's burm_state/1.
emit_state(S) ->
    Xs = to_string(duplicate(length(S#eburg.nts), x)),
    say("%% burm_state(BurmNode: node) -> #burm_state"),
    say("%% returns the state label for `BurmNode'."),
    say("burm_state(BurmNode) ->"),
    say("    InitC = { ~s },", [Xs]),
    say("    InitR = { ~s },", [Xs]),
    say("    case op_label(BurmNode) of"),
    emit_cases(S),
    say("    end,"),
    say("    #burm_state{cost = FinalC, rule = FinalR}.").

%% emit_cases(S: #eburg) -> ok
%% returns `ok' after printing the body of the case statement in `burm_state'.
emit_cases(S) ->
    %% for each operator
    sayforeach(fun ({Op, Rs}) ->
                       say("~s ->", [Op]),
                       if Rs == [] ->
                               say("burm_assert(length(kids(BurmNode)) == ~p),",
                                   [assoc(Op, S#eburg.arities)]),
                               say("{FinalC, FinalR} = {InitC, InitR}");
                          true ->
                               %% for each of its rules
                               reset(),
                               emit_rules(Rs, S, "InitC", "InitR")
                       end
               end,
               ";",
               [{Op, assoc(Op, S#eburg.oprules)} || Op <- S#eburg.ops]).

%% emit_closures(S: #eburg) -> ok
%% returns `ok' after printing one closure routine for every nonterminal
%% which has chainrules.
emit_closures(S) ->
    %% for each nonterminal
    sayforeach(fun ({Nt, Rs}) ->
                       say("%% burm_closure_~s(Node: tree, Cost: cost, Rule: rule, Base: int) -> {cost, rule}", [Nt]),
                       say("%% returns `Cost' and `Rule' after applying chainrules for ~s.", [Nt]),
                       say("burm_closure_~s(BurmNode, InitC, InitR, InitB) ->",
                           [Nt]),
                       %% for each of its chainrules
                       reset(),
                       emit_closure_body(Rs, S, "InitC", "InitR")
               end,
               "",
               [{Nt, assoc(Nt, S#eburg.crules)} ||
                   Nt <- S#eburg.nts, assoc(Nt, S#eburg.crules) /= []]).

%% emit_rules(Rs: [#rule], S: #eburg, CurC: string, CurR: string) -> ok
%% returns `ok' after printig the rules for a single case clause.
%% `CurC' and `CurR' are the names of the variables holding the most recent
%% cost/rule vectors.
%% XXX: rewrite as foldl?
emit_rules([], _S, CurC, CurR) ->
    say("{FinalC, FinalR} = {~s, ~s}", [CurC, CurR]);
emit_rules([R|Rs], S, CurC, CurR) ->
    say("%% ~s <- ~s", [R#rule.lhs, to_string(R#rule.rhs)]),
    {NewC, NewR} = emit_rule(R, S, CurC, CurR),
    say(","),
    emit_rules(Rs, S, NewC, NewR).

%% emit_closure_body(Rs: [#rule], S: #eburg, CurC: string, CurR: string) -> ok
%% returns `ok' after printing the code for one closure routine.
emit_closure_body([], _S, CurC, CurR) ->
    say("{~s, ~s}.", [CurC, CurR]);
emit_closure_body([R|Rs], S, CurC, CurR) ->
    say("%% ~s <- ~s", [R#rule.lhs, to_string(R#rule.rhs)]),
    {NewC, NewR} = emit_record(R, S, CurC, CurR, "InitB"),
    say(","),
    emit_closure_body(Rs, S, NewC, NewR).

%% emit_rule(R: #rule, S: #eburg, C: string, R: string) -> {string, string}
%% returns the names of the variables holding the latest cost and rule vectors
%% after printing code for handling (matching & recording)
%% a single rule.
%% FIXME leaves
emit_rule(R = #rule{rhs=#pattern{arity=0}}, S, CurC, CurR) ->
    %% no children, matching done by case pattern
    B0 = fresh_base(),
    say("begin"),
    say("    ~s = 0,", [B0]),
    {NewC, NewR} = emit_record(R, S, CurC, CurR, B0),
    say("end"),
    {NewC, NewR};
emit_rule(R = #rule{rhs=#pattern{kids=Kids, arity=N}}, S, CurC, CurR) ->
    say("burm_assert(length(kids(BurmNode)) == ~p),", [N]),
    B0 = fresh_base(),
    say("case ~s of", [gen_match(Kids, S)]),
    say("    true ->"),
    say("        ~s = ~s,", [B0, gen_cost(Kids, S)]),
    {NewC, NewR} = emit_record(R, S, CurC, CurR, B0),
    say(";"),
    say("    false ->"),
    say("        {~s, ~s} = {~s, ~s}", [NewC, NewR, CurC, CurR]),
    say("end"),
    {NewC, NewR}.

%% emit_record(R: #rule, S: #eburg, C: string, R: string, B: string) ->
%%                                                              {string, string}
%% returns the names of the variables holding the latest cost and rule vectors
%% after printing the code for recording rule numbers and costs.
%% It is assumed that the variable BurmNode points to the current node.
%% XXX: user macro names for rule numbers
emit_record(#rule{number=N, lhs=Nt, cost=Cost}, S, C0, R0, B0) ->
    if is_integer(Cost) -> RC = Cost;
       true -> RC = subst_pseudo_vars(Cost) % dynamic cost
    end,
    B1 = fresh_base(),
    C1 = fresh_cost(),
    R1 = fresh_rule(),
    M  = to_string({macref, Nt}),
    say("~s = ~s + ~p,", [B1, B0, RC]),
    say("case ~s  < element(~s, ~s) of", [B1, M, C0]),
    say("    true ->"),
    say("        ~s = setelement(~s, ~s, ~s),", [C1, M, C0, B1]),
    say("        ~s = setelement(~s, ~s, ~p),",  [R1, M, R0, N]),
    case assoc(Nt, S#eburg.crules) of
        [] ->
            {NewC, NewR} = {C1, R1},
            say("{~s, ~s};", [C1, R1]); % no closure routine
        _ ->
            C2 = fresh_cost(),
            R2 = fresh_rule(),
            {NewC, NewR} = {C2, R2},
            say("{~s, ~s} = burm_closure_~s(BurmNode, ~s, ~s, ~s);",
                [C2, R2, Nt, C1, R1, B1])
    end,
    say("    false ->"),
    say("        {~s, ~s} = {~s, ~s}", [NewC, NewR, C0, R0]),
    say("end"),
    {NewC, NewR}.

%%%_    . tree access

%%-----------------------------------------------------------------------------
%% `burm_state' needs to access the subtrees of the input tree rooted at the
%% current node to do tree pattern matching and cost calculations.
%% The functions in this section take a list of tree patterns from the RHS of
%% some rule, convert those patterns to a format understood by `emit_child'
%% and wrap some accessor and/or test code around the generated child
%% references.
%% It is assumed that the variable `BurmNode' points to the current node.
%%-----------------------------------------------------------------------------

%% gen_cost(Pats: [#pattern], S: #eburg) -> string
%% returns code to sum the costs of deriving the nonterminals in `Pats'.
gen_cost(Pats, S) ->
    saycat(fun ({Nt, Idx}) ->
                   ssay("?burm_cost(~s, ~s)",
                        [to_string({macref, Nt}), gen_child(Idx)])
           end,
           " + ",
           [{Nt, Idx} || {Nt, Idx} <- pats_to_idxs(Pats),
                         member(Nt, S#eburg.nts)]).

%% gen_match(Pats: [#pattern], S: #eburg) -> string
%% returns code to match `Pats' against a subject tree.
gen_match(Pats, S) ->
    saycat(fun ({OpOrNt, Idx}) ->
                   case member(OpOrNt, S#eburg.nts) of
                       true  ->
                           ssay("?burm_rule(~s, ~s) /= x",
                                [to_string({macref, OpOrNt}), gen_child(Idx)]);
                       false ->
                           ssay("op_label(~s) == ~s",
                                [gen_child(Idx), OpOrNt])
                   end
           end,
           " andalso ",
           pats_to_idxs(Pats)).

%% pats_to_idxs(Pats: [#pattern]) -> [{atom, [int]}]
%% returns the operators occuring in `Pats' paired with their positions in the
%% forest described by `Pats'.
%% An index of [3,2,1] is read as ``the third child of the second child
%% of the first child of the current node''.
pats_to_idxs(Pats)                   -> ps2i(Pats, [1]).
ps2i([Pat], Idx)                     -> p2i(Pat, Idx);
ps2i([Pat|Pats], Idx = [N|Ns])       -> p2i(Pat, Idx) ++ ps2i(Pats, [N+1|Ns]).
p2i(#pattern{op=Op, kids=[]},   Idx) -> [{Op, Idx}];
p2i(#pattern{op=Op, kids=Kids}, Idx) -> [{Op, Idx}] ++ ps2i(Kids, [1|Idx]).

%% gen_child(Idx: [int]) -> string
%% returns code to access the `Idx'th of the current node's children.
gen_child([])     -> "BurmNode";
gen_child([N|Ns]) -> ssay("lists:nth(~p, kids(~s))", [N, gen_child(Ns)]).

%%%_  . Reducer

%%-----------------------------------------------------------------------------
%% eburg provides several hardcoded reducers and some support routines for user
%% written reducers in the ``burm.hrl'' header file.
%% These rely on the following functions:
%%   o `nts' which takes an IRN and returns the INNs of the nonterminals
%%     occuring in that rule in the correct order
%%   o `burm_action' which is used by the hardcoded reducers to execute
%%      semantic action code
%%   o `is_chainrule' which is used by BURM's reducer
%%   o `burm_string' which maps IRNs and INNs to human-readable strings
%%-----------------------------------------------------------------------------

%% emit_reducers(S: #eburg) -> ok
%% returns `ok' after printing the reducer support routines.
emit_reducers(S) ->
    emit_is_chain(S),
    emit_burm_string(S),
    emit_nts(S),
    emit_action(S).

%% emit_nts(S: #eburg) -> ok
%% returns `ok' after writing `nts' to the output file.
emit_nts(S) ->
    say("%% nts(IRN: int) -> [int]"),
    say("%% returns the INNs of the nonterminals occuring in rule `IRN'."),
    say("nts(x) -> [];"), % XXX
    sayforeach(fun(R) -> say("nts(~p) -> ~s",[R#rule.number,gen_nts(R,S)]) end,
               ";",
               S#eburg.rules),
    say(".").

%% gen_nts(R: #rule, S: #eburg) -> string
%% returns code for a list containing the INNs of the nonterminals on the
%% RHS of `R' in the order in which they appear on the RHS of `R'.
gen_nts(R, S) ->
    ssay("[ ~s ]",
         [saycat(fun (Nt) -> to_string({macref, Nt}) end,
                 ", ",
                 nonterms(R, S))]).

%% nonterms(R: #rule, S: #eburg) -> [atom]
%% returns the nonterminals in the order in which they appear in `R'.
nonterms(#rule{rhs=Pat}, S)        -> nonterms([Pat], S);
nonterms([], _S)                   -> [];
nonterms([Pat|Pats], S)            -> nonterms(Pat, S) ++ nonterms(Pats, S);
nonterms(#pattern{kids=[K|Ks]}, S) -> nonterms(K, S) ++ nonterms(Ks, S);
nonterms(#pattern{op=OpOrNt, kids=[]}, S) ->
    case member(OpOrNt, S#eburg.nts) of
        true  -> [OpOrNt];
        false -> []
    end.

%% emit_action(S: #eburg) -> ok
%% returns `ok' after printing the semantic action routine.
emit_action(S) ->
    say("%% burm_action(Node: node, ReducedKids: [node], Rule: int) -> result"),
    say("%% returns the result of executing the semantic action associated"),
    say("%% with `Rule'."),
    say("burm_action(BurmNode, BurmReducedKids, BurmRule) ->"),
    say("case BurmRule of"),
    sayforeach(fun (R) ->
                       say("~p ->", [R#rule.number]),
                       io:put_chars(get('$out_stream'),
                                    subst_pseudo_vars(R#rule.action))
               end,
               ";",
               S#eburg.rules),
    say("end.").

%% emit_is_chain(S: #eburg) -> ok
%% returns `ok' after printing `is_chainrule'.
emit_is_chain(S) ->
    say("%% is_chainrule(IRN: int) -> bool"),
    say("%% returns `true' if rule number IRN is a chainrule."),
    sayforeach(fun (R) ->
                       say("is_chainrule(~p) -> ~p", 
                           [R#rule.number, is_chainrule(R, S)])
               end,
               ";",
               S#eburg.rules),
   say(".").

%% is_chainrule(R: #rule, S: #eburg) -> bool
%% returns `true' if `R' is a chainrule.
is_chainrule(#rule{rhs=#pattern{op=Op}}, #eburg{nts=Nts}) -> member(Op, Nts).

%% emit_burm_string(S: #eburg) -> ok
%% returns `ok' after printing `burm_string'.
emit_burm_string(S) ->
    say("%% burm_string({Tag: atom, IN: int}) -> string"),
    say("%% returns the name of `Tag' number `IN'."),
    say("burm_string({rule, x}) -> \"none\";"),
    sayforeach(fun (R) ->
                       say("burm_string({rule, ~p}) -> ~p", 
                           [R#rule.number, to_string(R)])
               end,
               ";",
               S#eburg.rules),
    say(";"),
    sayforeach(fun (Nt) ->
                       say("burm_string({nt, ~p}) -> ~p", 
                           [assoc(Nt, S#eburg.nt_nums), to_string(Nt)])
               end,
               ";",
               S#eburg.nts),

    say(".").

%%%_ , Utilities

%%%_  . Error handling

%%-----------------------------------------------------------------------------
%% `die' and `warning' bundle eburg error descriptors with context information
%% such as line numbers and throw the result as an eburg_* exception.
%% These exceptions are caught in two places: `file' catches eburg_error
%% exceptions (so `die' effectively halts the computation) and `check'
%% catches eburg_warning exceptions, prints an appropriate message and
%% continues. `format_error' maps error_descriptors to strings.
%%-----------------------------------------------------------------------------

%% die(File: string, Reason: error_descriptor)
%% die(File: string, Line: int, Reason: error_descriptor)
%% throws an `eburg_error' exception.
die(File, Reason)       -> throw({eburg_error, File, Reason}).
die(File, Line, Reason) -> throw({eburg_error, File, Line, Reason}).

%% warning(File: string, Reason: error_descriptor)
%% warning(File: string, Line: int, Reason: error_descriptor)
%% throws an `eburg_warning' exception.
warning(File, Reason)       -> throw({eburg_warning, File, Reason}).
warning(File, Line, Reason) -> throw({eburg_warning, File, Line, Reason}).

%% err(File: string, Msg: string) -> ok
%% err(File: string, Line: int, Msg: string) -> ok
%% returns `ok' after printing an Emacs-friendly error message.
err(File, Msg) ->
    io:format("~s: ~s", [File, Msg]).
err(File, Line, Msg)  ->
    io:format("~s:~w: ~s", [File, Line, Msg]).

%% format_error(Reason: error_descriptor) -> string
%% returns a descriptive error message for `Reason'.
%% FIXME
format_error({file, Reason}) ->
    file:format_error(Reason);
format_error({syntax, {_, Module, Reason}}) ->
    apply(Module, format_error, Reason);
format_error(uniq) ->
    "multiple declarations of the same operator";
format_error(diff) ->
    "operator and nonterminal names overlap";
format_error(start) ->
    "no rules for declared start symbol";
format_error({undeclared, Op}) ->
    atom_to_list(Op) ++ " used but not declared";
format_error({arity, Op}) ->
    atom_to_list(Op) ++ " used with different arities";
format_error({unused, Ops}) ->
    "unused operator" ++
        if length(Ops) > 1 -> "s: ";
           true -> ": "
        end ++ to_string(Ops);
format_error({unreachable, Nts}) ->
    "unreachable nonterminal" ++
        if length(Nts) > 1 -> "s: ";
           true -> ": "
        end ++ to_string(Nts).

%%%_  . Strings and I/O

%%-----------------------------------------------------------------------------
%% `say' is a convenience wrapper around for `io:format' which always prints
%% to ```get('$out_stream')''' (which must resolve to an `io_device').
%% `sayforeach' is somewhat similar to `lists:foreach'; it assumes that its
%% first argument (usually a fun which calls `say') applied to the elements of
%% its third argument will print some Erlang code which should be separated by
%% printing its second argument (typically ",", ";" or somesuch).
%%
%% `ssay' and `saycat' are like `say' and `sayforeach' but for in-memory
%% strings rather than I/O.
%%-----------------------------------------------------------------------------

%% say(Fmt: string) -> ok
%% say(Fmt: string, Data: [term]) -> ok
%% returns `ok' after printing `Data' formatted according to `Fmt'.
say(Fmt)       -> io:format(get('$out_stream'), Fmt ++ "~n", []).
say(Fmt, Data) -> io:format(get('$out_stream'), Fmt ++ "~n", Data).

%% sayforeach(Say: fun, Sep: string, Xs: [term]) -> ok
%% returns `ok' after calling `Say' to print all the elements of `Xs' in turn,
%% injecting `Sep' after each element.
sayforeach(  _,   _, [])     -> ok;
sayforeach(Say,   _, [X])    -> Say(X);
sayforeach(Say, Sep, [X|Xs]) -> Say(X), say(Sep), sayforeach(Say, Sep, Xs).

%% ssay(Fmt: string, Data: [term]) -> string
%% returns Data formatted according to Fmt.
ssay(Fmt, Data) -> flatten(io_lib:format(Fmt, Data)).

%% saycat(Say: fun, Sep: string, Xs: [term]) -> string
%% returns the result of linking the results of calling `Say' on the elements
%% of `Xs' by `Sep'.
saycat(  _,   _, [])     -> "";
saycat(Say,   _, [X])    -> Say(X);
saycat(Say, Sep, [X|Xs]) -> Say(X) ++ Sep ++ saycat(Say, Sep, Xs).

%% to_string(Obj: printable_object) -> string
%% returns the string representation of `Obj'.
to_string([]) ->
    "";
to_string([X]) ->
    to_string(X);
to_string([X|Xs]) ->
    to_string(X) ++ ", " ++ to_string(Xs);
to_string(X) when is_atom(X) ->
    atom_to_list(X);
to_string({macdef, Nt}) when is_atom(Nt) ->
    map(fun to_upper/1, atom_to_list(Nt));
to_string({macdef, R}) when is_record(R, rule) ->
    Prefix      = to_string(R#rule.lhs) ++ "_",
    {ok, S0, _} = regexp:sub(to_string(R#rule.rhs), "(\)+)$", ""),
    {ok, S1, _} = regexp:gsub(S0, "[(,)]", "_"),
    {ok, S2, _} = regexp:gsub(S1, "__", "_"),
    {ok, S3, _} = regexp:gsub(S2, " ", ""),
    S4          = Prefix ++ S3,
    map(fun to_upper/1, S4);
to_string({macref, Mac}) ->
    "?" ++ to_string({macdef, Mac});
to_string(#rule{lhs=LHS, rhs=RHS}) ->
    to_string(LHS) ++ " <- " ++ to_string(RHS);
to_string(#pattern{op=Op, kids=[]}) ->
    to_string(Op);
to_string(#pattern{op=Op, kids=Kids}) ->
    to_string(Op) ++ "(" ++ to_string(Kids) ++ ")".

%% to_upper(C: char) -> char
%% returns the upper case version of `C'.
to_upper(C) ->
    if ($a =< C) andalso (C =< $z) -> C - ($a - $A);
       true -> C
    end.

%%%_  . Temp and pseudo vars

%%-----------------------------------------------------------------------------
%% `fresh' and `current' implement a simple gensym-like facility for strings.
%%-----------------------------------------------------------------------------

%% fresh(Base: string) -> string
%% returns a fresh name formed by appending an integer to `Base'.
fresh(Base) ->
    Sym = [$$|Base],
    N   =
        case get(Sym) of
            undefined -> 0;
            Old       -> Old + 1
        end,
    put(Sym, N),
    Base ++ integer_to_list(N).

%% current(Base: string) -> string | undefined
%% returns the last name returned by `fresh'.
current(Base) ->
    case get([$$|Base]) of
        undefined -> undefined;
        N         -> Base ++ integer_to_list(N)
    end.

%% reset() -> ok
%% returns `ok' after clearing the process dictionary.
reset() ->
    erase("$Cost"),
    erase("$Rule"),
    erase("$Base"),
    ok.

%% fresh_[cost|rule|base]() -> atom
%% returns a previously unused intermediate variable name.
fresh_cost() -> fresh("Cost").
fresh_rule() -> fresh("Rule").
fresh_base() -> fresh("Base").

%% current_[cost|rule|base]() -> atom
%% returns the last used intermediate variable name.
current_cost() -> current("Cost").
current_rule() -> current("Rule").
current_base() -> current("Base").

%%-----------------------------------------------------------------------------
%% `$a' can occur in dynamic cost functions or semantic actions. It expands to
%% the current node (`BurmNode' in the scope of code emitted by `emit_record'
%% and `emit_reducer').
%%
%% `$N' where `N' is some integer in [0-9] should occur only in semantic actions
%% where it expands to the result of reducing the Nth child of the current node
%% (the reduced children are kept in `BurmReducedKids' in the scope of code
%% emitted by `emit_reducer').
%%-----------------------------------------------------------------------------

%% subst_pseudo_vars(Code: string) -> string
%% returns `Code' with pseudo variables expanded.
subst_pseudo_vars("") ->
    "";
subst_pseudo_vars([$\', $$, $a, $\'|Rest]) ->
    "BurmNode" ++ subst_pseudo_vars(Rest);
subst_pseudo_vars([$\', $$, N, $\'|Rest]) ->
    "lists:nth(" ++ [N] ++ ", BurmReducedKids)" ++ subst_pseudo_vars(Rest);
subst_pseudo_vars([First|Rest]) ->
    [First|subst_pseudo_vars(Rest)].

%%%_  . Misc

%% pp(File: name) -> ok
%% returns `ok' after pretty-printing `File'.
pp(File) -> erl_tidy:file(File, [{backups, false}, idem]).

%% mk_reader(Lexer: fun, Parser: fun) -> fun
%% returns a function which uses `Lexer' and `Parser' (assumed to follow yecc
%% conventions) to tokenize and parse its single argument (must be a string).
mk_reader(Lexer, Parser) ->
    fun (String) ->
            case Lexer(String) of
                {error, Reason, Line} -> {error, Reason, Line};
                {ok, Tokens, _} ->
                    case Parser(Tokens) of
                        {error, Reason, Line} -> {error, Reason, Line};
                        {ok, Result} -> Result
                    end
            end
    end.

%% assoc(Key: term, Alist: [{term, term}]) -> term
%% returns the value associated with `Key' in `Alist'.
assoc(Key, Alist) -> {Key, Val} = proplists:lookup(Key, Alist), Val.

%%%_. Emacs

%%% Local Variables:
%%% mode: erlang
%%% allout-layout: t
%%% End:
